@@ -1,7 +1,8 @@
 """pypyr pipeline yaml definition classes - domain specific language."""
-
+import json
 import logging
-import os
+from ruamel.yaml.comments import CommentedMap, CommentedSeq
+from ruamel.yaml.nodes import ScalarNode
 from pypyr.errors import (Call,
                           ControlOfFlowInstruction,
                           get_error_name,
@@ -15,10 +16,6 @@ from pypyr.utils import expressions, poll
 # use pypyr logger to ensure loglevel is set correctly
 logger = logging.getLogger(__name__)
 
-
-# os.getenv's dict fills on 1st import of os.
-skip_clean_in_on_step_done = (os.getenv('PYPYR_IN_CLEAN', '0')
-                              not in {'true', 'TRUE', '1'})
 # ------------------------ custom yaml tags -----------------------------------
 
 
@@ -50,7 +47,7 @@ class SpecialTagDirective:
 
     def __str__(self):
         """Friendly string representation of instance."""
-        return self.value
+        return str(self.value)
 
     def __repr__(self):
         """Repr representation of instance."""
@@ -84,6 +81,92 @@ class SpecialTagDirective:
             'during formatting operations.')
 
 
+class Jsonify(SpecialTagDirective):
+    """Serialize context structure into a json string.
+
+    Runs formatting substitution expressions on all formattable fields in the
+    input structure.
+    """
+
+    yaml_tag = '!jsonify'
+
+    def __init__(self, value, scalar=None):
+        """Initialize class with special handling for TaggedScalar.
+
+        Args:
+            value (any): Value of yaml object. Likely CommentedMap or
+                         CommentedSeq or if scalar the mapped Python type for
+                         the scalar.
+            scalar (TaggedScalar): If type scalar, the original constructed
+                                   object. This is necessary for to_yaml
+                                   serialization.
+        """
+        self.scalar = scalar
+        self.value = value
+
+    def __repr__(self):
+        """Handle TaggedScalar specially.
+
+        This is because original node necessary to reconstruct a TaggedScalar
+        and .value gives the underlying simple type value instead.
+        """
+        if self.scalar:
+            return (
+                f'{self.__class__.__name__}({self.value!r}, {self.scalar!r})')
+        else:
+            return f'{self.__class__.__name__}({self.value!r})'
+
+    @classmethod
+    def from_yaml(cls, constructor, node):
+        """Create the class from yaml representation."""
+        for data in constructor.construct_undefined(node):
+            # returns generator as a means to update the
+            # returned iterator for recursive yaml refs where the node is still
+            # under construction.
+            pass
+
+        if isinstance(node, ScalarNode):
+            if node.style:
+                # style having a value means it is quoted, i.e always a str
+                # therefore don't need to go and hunt down the type
+                return cls(data.value, data)
+
+            # constructed_undefined creates all scalar as TaggedScalar, which
+            # is always str. Use resolver to construct object to get it to
+            # parse to the appropriate Python literal simple type.
+            tag = constructor.resolver.resolve(ScalarNode,
+                                               node.value,
+                                               (True, False))
+
+            scalar_node = ScalarNode(tag,
+                                     node.value,
+                                     start_mark=node.start_mark,
+                                     end_mark=node.end_mark,
+                                     style=node.style,
+                                     comment=node.comment,
+                                     anchor=node.anchor)
+
+            constructed_scalar_node = constructor.construct_object(scalar_node)
+            return cls(constructed_scalar_node, data)
+
+        return cls(data)
+
+    @classmethod
+    def to_yaml(cls, representer, node):
+        """Serialize this class back to yaml."""
+        if isinstance(node.value, CommentedMap):
+            return representer.represent_mapping(cls.yaml_tag, node.value)
+        elif isinstance(node.value, CommentedSeq):
+            return representer.represent_sequence(cls.yaml_tag, node.value)
+        else:
+            return representer.represent_tagged_scalar(node.scalar)
+
+    def get_value(self, context):
+        """Serialize self contents to json."""
+        # dumps writes null if self.value is None
+        return json.dumps(context.get_formatted_value(self.value))
+
+
 class PyString(SpecialTagDirective):
     """py strings allow you to execute python expressions dynamically.
 
@@ -275,10 +358,6 @@ class Step:
 
         # description: optional. Write to stdout if exists and flagged.
         self.description = step.get('description', None)
-        if self.description:
-            logger.notify("%s: %s", self.name, self.description)
-        else:
-            logger.debug("step name: %s", self.name)
 
         # foreach: optional value. None by default.
         self.foreach_items = step.get('foreach', None)
@@ -308,6 +387,8 @@ class Step:
         if while_definition:
             self.while_decorator = WhileDecorator(while_definition)
 
+        logger.debug("step name: %s", self.name)
+
     def save_error(self, context, exception, swallowed):
         """Append step's exception information to the context.
 
@@ -324,7 +405,7 @@ class Step:
         failure = {
             'name': get_error_name(exception),
             'description': str(exception),
-            'customError': context.get_formatted_iterable(
+            'customError': context.get_formatted_value(
                 self.on_error
             ) if self.on_error else {},
             'line': self.line_no,
@@ -352,7 +433,7 @@ class Step:
 
         # Loop decorators only evaluated once, not for every step repeat
         # execution.
-        foreach = context.get_formatted_iterable(self.foreach_items)
+        foreach = context.get_formatted_value(self.foreach_items)
 
         foreach_length = len(foreach)
 
@@ -406,7 +487,9 @@ class Step:
             finally:
                 self.reset_context_counters(context, call)
 
-            logger.debug("call: done calling %s", call.groups)
+            # py 3.9 issue with coveragepy doesn't show this as covered. it is.
+            logger.debug("call: done calling %s",
+                         call.groups)  # pragma: no cover
 
         logger.debug("step %s done", self.name)
 
@@ -557,6 +640,19 @@ class Step:
                      mutate.
         """
         logger.debug("starting")
+
+        # give user helpful output if step will actually run or not.
+        if self.description:
+            description = context.get_formatted_value(self.description)
+            run_me = context.get_formatted_as_type(self.run_me, out_type=bool)
+            skip_me = context.get_formatted_as_type(self.skip_me,
+                                                    out_type=bool)
+
+            if run_me and not skip_me:
+                logger.notify(description)
+            else:
+                logger.notify("(skipping): %s", description)
+
         # the in params should be added to context before step execution.
         self.set_step_input_context(context)
 
@@ -608,14 +704,8 @@ class Step:
         """
         logger.debug("starting")
 
-        if skip_clean_in_on_step_done:
-            logger.debug("skipping 'in' removal after step because "
-                         "$PYPYR_IN_CLEAN is not set or 0.")
-            logger.debug("done")
-            return
-
         if self.in_parameters is not None:
-            # len is O(1), so not all that unecessarily duplicated cpu time
+            # len is O(1), so not all that unnecessarily duplicated cpu time
             parameter_count = len(self.in_parameters)
             if parameter_count > 0:
                 logger.debug(
@@ -739,7 +829,7 @@ class RetryDecorator:
             if self.stop_on or self.retry_on:
                 error_name = get_error_name(ex_info)
                 if self.stop_on:
-                    formatted_stop_list = context.get_formatted_iterable(
+                    formatted_stop_list = context.get_formatted_value(
                         self.stop_on)
                     if error_name in formatted_stop_list:
                         logger.error("%s in stopOn. Raising error "
@@ -749,7 +839,7 @@ class RetryDecorator:
                         logger.debug("%s not in stopOn. Continue.", error_name)
 
                 if self.retry_on:
-                    formatted_retry_list = context.get_formatted_iterable(
+                    formatted_retry_list = context.get_formatted_value(
                         self.retry_on)
                     if error_name not in formatted_retry_list:
                         logger.error("%s not in retryOn. Raising "
