@@ -36,6 +36,7 @@ def run_step(context):
         key4: value3
     """
     logger.debug("started")
+    deprecated(context)
     context.assert_key_has_value(key='contextSet', caller=__name__)
 
     for k, v in context['contextSet'].items():
@@ -43,3 +44,11 @@ def run_step(context):
         context[k] = context[v]
 
     logger.debug("done")
+
+
+def deprecated(context):
+    """Handle deprecated context input."""
+    logger.warning("contextSet is deprecated. It will "
+                   "stop working upon the next major release. "
+                   "Use the new pypyr.steps.contextcopy instead. It's a lot "
+                   "better, promise!")
